structure CeXApp =
  struct
    structure Draw = Drawing
    structure Geom = Geometry
    structure I = Interact
    structure EXB = EXeneBase;
    structure cam = ce3DCamera;
    structure vec = ce3DVector;
    structure vert = ce3DVertex;
    structure matr = ce3DMatrix;
    structure poly = ce3DPoly;
    structure ang  = ce3DAngle;
    structure rect = ceRect;
    structure wren = ceWireRender;

    type appData = { angleX : int,
                     angleY : int,
                     angleZ : int,
                     shearX : real,
                     shearY : real,
                     shearZ : real,
                     scale : real,
                     camera : cam.camera,
                     sphere : poly.polygonMesh };

    (* Obsolete data... *)
    val VRP = (0.0, 0.0, 20.0);
    val VPN = (0.0, 0.0, 1.0);
    val VUP = (0.0, 1.0, 0.0);
    val PRP = (0.0, 0.0, 30.0);
    val viewPlane = ((~9.0, ~6.6), (18.0, 13.2));(* viewPlane and viewPort must be proportional *)
    val viewPort = ((10.0, 10.0), (600.0, 440.0));
    val (frontClipPlane, backClipPlane) = (0.0, ~40.0);

    fun makeAppSphere () =
          let
            val (meshVert, meshLines, meshPoly) =
                  ce3DPrimitiveGenerators.generateSphere
                    (20,20,3.0);
            val placeMatr = matr.translateXYZMatrix(0.0, 3.0, 0.0);
            val transFnc = matr.transformVertexFunction placeMatr;
            fun foldTFunc (elem, list) =
                  (transFnc elem)::list
            val transVert =
                  Array.fromList
                    (Array.foldr foldTFunc nil meshVert);
          in
            (transVert, meshLines, meshPoly)
          end
    val gridMesh = ce3DPrimitiveGenerators.generateGrid(20,20);

    fun newApp () =
      { angleX = 0,
        angleY = 0,
        angleZ = 0,
        shearX = 0.0,
        shearY = 0.0,
        shearZ = 0.0,
        scale  = 100.0,
        camera = cam.defaultCamera(),
        sphere = makeAppSphere() }

    fun rotX ({angleX = aX,angleY,angleZ,shearX,shearY,shearZ,scale,camera,sphere} : appData, angle : int) =
      { angleX = (aX + angle), angleY = angleY, angleZ = angleZ,
        shearX = shearX, shearY = shearY, shearZ = shearZ,
        scale = scale, camera = camera, sphere = sphere}
    
    fun rotY ({angleX,angleY=aY,angleZ,shearX,shearY,shearZ,scale,camera,sphere} : appData, angle : int) =
      { angleX = angleX, angleY = (aY + angle), angleZ = angleZ,
        shearX = shearX, shearY = shearY, shearZ = shearZ,
        scale = scale, camera = camera, sphere = sphere }

    fun rotZ ({angleX,angleY,angleZ=aZ,shearX,shearY,shearZ,scale,camera,sphere} : appData, angle : int) =
      { angleX = angleX, angleY = angleY, angleZ = (aZ + angle),
        shearX = shearX, shearY = shearY, shearZ = shearZ,
        scale = scale, camera = camera, sphere = sphere }

    fun shearX ({angleX,angleY,angleZ,shearX=shX,shearY,shearZ,scale,camera,sphere} : appData, shear : real) =
      { angleX = angleX, angleY = angleY, angleZ = angleZ,
        shearX = (shX + shear), shearY = shearY, shearZ = shearZ,
        scale = scale, camera = camera,sphere = sphere}

    fun shearY ({angleX,angleY,angleZ,shearX,shearY=shY,shearZ,scale,camera,sphere} : appData, shear : real) =
      { angleX = angleX, angleY = angleY, angleZ = angleZ,
        shearX = shearX, shearY = (shY + shear), shearZ = shearZ,
        scale = scale, camera = camera, sphere = sphere}

    fun shearZ ({angleX,angleY,angleZ,shearX,shearY,shearZ=shZ,scale,camera,sphere} : appData, shear : real) =
      { angleX = angleX, angleY = angleY, angleZ = angleZ,
        shearX = shearX, shearY = shearY, shearZ = (shZ + shear),
        scale = scale, camera = camera, sphere = sphere}

    fun scale ({angleX,angleY,angleZ,shearX,shearY,shearZ,scale=sca,camera, sphere} : appData, scale : real) =
      { angleX = angleX, angleY = angleY, angleZ = angleZ,
        shearX = shearX, shearY = shearY, shearZ = shearZ,
        scale = (sca + scale), camera = camera, sphere = sphere }

    fun dolly ({angleX,angleY,angleZ,shearX,shearY,shearZ,scale,camera, sphere} : appData, dist : real) =
      { angleX = angleX, angleY = angleY, angleZ = angleZ,
        shearX = shearX, shearY = shearY, shearZ = shearZ,
        scale = scale, camera = cam.dolly(camera,dist), sphere = sphere }

    fun tumbleX ({angleX,angleY,angleZ,shearX,shearY,shearZ,scale,camera, sphere} : appData, rotAngle : ang.angle) =
      { angleX = angleX, angleY = angleY, angleZ = angleZ,
        shearX = shearX, shearY = shearY, shearZ = shearZ,
        scale = scale, camera = cam.tumbleX(camera,rotAngle), sphere = sphere }

    fun tumbleY ({angleX,angleY,angleZ,shearX,shearY,shearZ,scale,camera, sphere} : appData, rotAngle : ang.angle) =
      { angleX = angleX, angleY = angleY, angleZ = angleZ,
        shearX = shearX, shearY = shearY, shearZ = shearZ,
        scale = scale, camera = cam.tumbleY(camera,rotAngle), sphere = sphere }

    fun trackX ({angleX,angleY,angleZ,shearX,shearY,shearZ,scale,camera, sphere} : appData, dist : real) =
      { angleX = angleX, angleY = angleY, angleZ = angleZ,
        shearX = shearX, shearY = shearY, shearZ = shearZ,
        scale = scale, camera = cam.trackX(camera,dist), sphere = sphere }

    fun trackY ({angleX,angleY,angleZ,shearX,shearY,shearZ,scale,camera, sphere} : appData, dist : real) =
      { angleX = angleX, angleY = angleY, angleZ = angleZ,
        shearX = shearX, shearY = shearY, shearZ = shearZ,
        scale = scale, camera = cam.trackY(camera,dist), sphere = sphere }

    fun renderWire (drawable, pen, dimPen, data : appData) =
      let
        val lines = wren.render(#camera data, viewPort,#sphere data);
        val gridLines = wren.render(#camera data, viewPort, gridMesh);
        val _ = Draw.clearDrawable drawable
        val _ = Draw.drawSegs drawable dimPen gridLines
        val _ = Draw.drawSegs drawable pen lines
      in
        ()
      end

    fun launchApp ( mouseFirstDownEvt,
                    mouseLastUpEvt, 
                    mouseMoveEvt,
                    cmdEvt,
                    mouseMoveCh,
                    mouseReplyCh,
                    drawable,
                    pen,
                    dimPen,
                    bgPen) =
      let
        fun loop (appdata,point,mb1,mb2,mb3) =
              let
                fun handleCmd (I.CI_Redraw _) =
                    let
                      val (Geom.PT{x, y}) = point;
                    in(* refresh screen *)
                      (renderWire(
                               drawable,
                               pen,
                               dimPen,
                               appdata);
                       loop (appdata,point,mb1, mb2, mb3))
                    end
                  | handleCmd I.CI_OwnDeath = ()
                  | handleCmd _ = (loop (appdata,point,mb1, mb2, mb3))
                fun firstDown (pt, I.MButton(1)) =
                      (loop(appdata,pt, true, mb2, mb3))
                  | firstDown (pt, I.MButton(2)) =
                      (loop(appdata, pt, mb1, true, mb3))
                  | firstDown (pt, I.MButton(3)) =
                      (loop(appdata, pt, mb1, mb2, true))
                  | firstDown (pt, _) =
                      (loop(appdata, point, mb1, mb2, mb3))
                fun lastUp (pt, I.MButton(1)) =
                      (loop(appdata,point, false, mb2, mb3))
                  | lastUp (pt, I.MButton(2)) =
                      (loop(appdata, point, mb1, false, mb3))
                  | lastUp (pt, I.MButton(3)) =
                      (loop(appdata, point, mb1, mb2, false))
                  | lastUp (pt, _) =
                      (loop(appdata, point, mb1, mb2, mb3))
                fun move pt =
                  let
                    fun pollMoves(inPoint) =
                      let
                        val _ = CML.yield();(* If this isn't called, polling will rarely occur *)
                        val polled = CML.recvPoll(mouseMoveCh);
                      in
                        case polled of
                          NONE =>
                            inPoint
                        | SOME(polledPoint) =>
                            pollMoves(polledPoint)
                      end
                    val (Geom.PT{x, y}) = pollMoves(pt);
                    val (Geom.PT{x = oldX, y = oldY}) = point;
                  in
                     if mb3 then
                       let
                         val appdata = dolly(appdata,Real.fromInt(x - oldX) / 10.0);
                       in
                         (renderWire(
                                   drawable,
                                   pen,
                                   dimPen,
                                   appdata);
                         CML.yield();(* Instead of flush *)
                         loop(appdata,pt,mb1, mb2, mb3))
                       end
                     else if mb1 then
                       let
                         val appdata = tumbleX(appdata, ang.fromInt(oldY - y) / 100.0);
                         val appdata = tumbleY(appdata, ang.fromInt(oldX - x) / 100.0);
                       in
                         (renderWire(
                                   drawable,
                                   pen,
                                   dimPen,
                                   appdata);
                         CML.yield();(* Instead of flush *)
                         loop(appdata,pt,mb1, mb2, mb3))
                       end
                     else if mb2 then
                       let
                         val appdata = trackX(appdata, ang.fromInt(x - oldX) / 50.0);
                         val appdata = trackY(appdata, ang.fromInt(oldY - y) / 50.0);
                       in
                         (renderWire(
                                   drawable,
                                   pen,
                                   dimPen,
                                   appdata);
                         CML.yield();(* Instead of flush *)
                         loop(appdata,pt,mb1, mb2, mb3))
                       end
                     else
                       (loop(appdata,point,mb1, mb2, mb3))
                  end
              in
                CML.select [
                    CML.wrap (cmdEvt, handleCmd),
                    CML.wrap (mouseFirstDownEvt, firstDown),
                    CML.wrap (mouseLastUpEvt, lastUp),
                    CML.wrap (mouseMoveEvt, move)
                  ]
              end
      in
        CML.spawn (fn () => loop (newApp(),Geom.PT{x=0,y=0},false,false,false));
        ()
      end (* launchApp *)

  end
;