structure gui : GUI =      
  struct
    structure W = Widget
    structure A = Attrs
    structure Draw = Drawing
    structure Geom = Geometry
    structure EXB = EXeneBase
    structure I = Interact;
    structure cexApp = CeXApp;
    
    val resources = ["*background: gray"]
    val minSz = Geom.SIZE{wid=640,ht=480}

    val maxcolor = 0w65535
    val medcolor = maxcolor div 0w2
    val quadcolor= medcolor div 0w2
    val mincolor = 0w0 

    val graycol = W.EXB.CMS_RGB{red=quadcolor+medcolor,
                                green=quadcolor+medcolor,
                                blue=quadcolor+medcolor}
    
    fun initTopWin dpyName =
      let
	val dpy = (EXB.openDisplay (dpyName, NONE))
		handle EXB.BadAddr s => (
		  TextIO.print s; TextIO.print "\n";
		  RunCML.shutdown OS.Process.failure)
        val scr = EXB.defaultScreenOf dpy
	val winSz = Geom.SIZE{wid=640,ht=480}
	val (win, inEnv) =
	      EXeneWin.createSimpleTopWin scr {
		  geom = Geom.WGEOM{pos=Geom.PT{x=0, y=0}, sz=winSz, border=1},
		  border = EXB.whiteOfScr scr,
		  backgrnd = EXB.blackOfScr scr
		}
	in
	  EXeneWin.setWMProperties win {
	      argv = SMLofNJ.getArgs(),
	      win_name = SOME "CeX 3D Test...",
	      icon_name = SOME "CeX 3D Test...",
	      size_hints = [
		  ICCC.HINT_PPosition,
		  ICCC.HINT_PSize,
		  ICCC.HINT_PMinSize minSz
		],
	      wm_hints = [],
	      class_hints = SOME{res_name="cex", res_class="CeX"}
	    };
	  EXeneWin.mapWin win;
          (scr, win, inEnv)
	end

    fun drawWinGeom (wid, ht) = Geom.WGEOM{
          pos = Geom.PT{x = 5, y = 5},
          sz = Geom.SIZE{wid = wid - 10, ht = ht - 10},
          border = 1
        }

    fun allocWindows dpyName =
      let
        val (scr,topWin,topEnv) = initTopWin dpyName
        val (Geom.SIZE{wid, ht}) = EXB.sizeOfWin topWin
        val drawWin = EXeneWin.createSimpleSubwin topWin
              {
                geom = drawWinGeom(wid, ht),
                border = SOME(EXB.whiteOfScr scr),
                backgrnd = SOME(EXB.blackOfScr scr)
              }
      in
        EXeneWin.mapWin drawWin;
        (scr, {top_win=topWin, top_env=topEnv, draw_win=drawWin})
      end

    fun mkTopLevelThreads {top_win, top_env=I.InEnv{k, m, ci, ...}, draw_win} =
      let
        val (drawInEnv, drawOutEnv) = I.createWinEnv()
        val (inEnv, outEnv) = I.createWinEnv()
        val inEnv = I.ignoreAll inEnv
        fun findEnv msg = (case (I.stripMsg msg)
               of (I.Here _) => outEnv
                | (I.ToChild msg') => (
                    if (I.toWindow(msg', draw_win))
                      then drawOutEnv
                    else raise (Fail "findEnv"))
              (* end case *))
        val kbdEvt = (CML.wrap(k, fn msg => let
                val I.OutEnv{k, ...} = findEnv msg
                in
                  CML.sync (k msg)
                end))
         val mouseEvt = (CML.wrap(m, fn msg => let
                val I.OutEnv{m, ...} = findEnv msg
                in
                  CML.sync (m msg)
                end))
         val cmdEvt = (CML.wrap(ci, fn msg => let
                val I.OutEnv{ci, ...} = findEnv msg
                in
                  CML.sync (ci msg)
                end))
         fun router () = (
              CML.select [kbdEvt, mouseEvt, cmdEvt];
              router())
       in
         CML.spawn router;
         { draw_env = drawInEnv}
       end
    
    fun mkEventThreads (scr, win, env) =
      let
        val I.InEnv{m, ci, ...} = I.ignoreKey env
        val mouseEvt = CML.wrap (m, I.msgBodyOf)
        val cmdEvt = CML.wrap (ci, I.msgBodyOf)
        val mouseFirstDownCh = CML.channel()
        val mouseLastUpCh    = CML.channel()
        val mouseMoveCh      = CML.channel()
        val mouseReplyCh     = CML.channel()

        val mouseFirstDownEvt = CML.recvEvt mouseFirstDownCh
        val mouseLastUpEvt    = CML.recvEvt mouseLastUpCh
        val mouseMoveEvt      = CML.recvEvt mouseMoveCh
        val mouseReplyEvt     = CML.recvEvt mouseReplyCh
        
        fun mousePrivateLoop () =
            (case (CML.select [mouseReplyEvt, mouseEvt])
               of I.MOUSE_FirstDown{pt, but, ...} => CML.send(mouseFirstDownCh, (pt, but) )
                | I.MOUSE_LastUp{pt, but, ...} => CML.send(mouseLastUpCh, (pt, but) )
                | I.MOUSE_Motion{pt,...} => CML.send(mouseMoveCh,pt)
                | _ => ()
              (* end case *);
            mousePrivateLoop())
        
        val drawable = Draw.drawableOfWin win
        val unBufDrawable = Draw.feedback drawable;
        val grayucol = W.EXB.colorOfScr scr graycol
        val pen = Draw.newPen
                [ Draw.PV_Function Draw.OP_Copy,
                  Draw.PV_Foreground(EXB.whiteOfScr(EXeneWin.screenOfWin win)),
                  Draw.PV_LineWidth 1
                ]
        val dimPen = Draw.newPen
                [ Draw.PV_Function Draw.OP_Copy,
                  Draw.PV_Foreground(grayucol),
                  Draw.PV_LineWidth 1
                ]
        val bgPen = Draw.newPen
                [
                  Draw.PV_Foreground(EXB.colorOfScr(EXeneWin.screenOfWin win) (EXB.CMS_RGB({red=Word.fromInt(200),green=Word.fromInt(200),blue=Word.fromInt(200)}))),
                  Draw.PV_LineWidth 1
                ]
      in
        CML.spawn mousePrivateLoop;
        (
          mouseFirstDownEvt,
          mouseLastUpEvt,
          mouseMoveEvt,
          cmdEvt,
          mouseMoveCh,
          mouseReplyCh,
          unBufDrawable,
          pen,
          dimPen,
          bgPen
        )
      end (* mkEventThreads *)
    
    fun threadInt dpyName =
      let
        open CML Interact
        val (src, x as {draw_win,...}) = allocWindows dpyName
        val {draw_env} = mkTopLevelThreads x
      in
        cexApp.launchApp(mkEventThreads(src, draw_win, draw_env))
      end

    fun doit () =
      let
        val _ = (
                  XDebug.init [];
                  RunCML.doit(
                    fn() => (XDebug.xspawn("threadInt",fn() => threadInt ""); ()),
                    SOME(Time.fromMilliseconds 20))
                )
      in
        ()
      end

    fun runArgs (arg0 : string, args : string list) =
      (
       XDebug.init [];
       RunCML.doit(
         fn() => (XDebug.xspawn("threadInt",fn() => threadInt ""); ()),
         SOME(Time.fromMilliseconds 20))
      )

  end
;