structure ce3DAdvMatrix : CE3DADVMATRIX =
  struct
    structure matr : CE3DMATRIX = ce3DMatrix;
    structure ang : CE3DANGLE = ce3DAngle;
    structure vert : CE3DVERTEX = ce3DVertex;
    structure vec : CE3DVECTOR = ce3DVector;

    (* Special viewing system matrices *)
    fun viewVectorsRotationMatrix (VPN : vec.t, VUP : vec.t) = (* Not tested! *)
      let
        val (nVPNX, nVPNY, nVPNZ) = vec.normalize(VPN);
        val cVUP = vec.crossProduct(VUP, (nVPNX, nVPNY, nVPNZ));
        val (cnVUPX, cnVUPY, cnVUPZ) = vec.normalize(cVUP);
        val (cX, cY, cZ) = vec.crossProduct((nVPNX, nVPNY, nVPNZ),(cnVUPX, cnVUPY, cnVUPZ));
      in
        matr.genericMatrix
          ((cnVUPX, cnVUPY, cnVUPZ, 0.0),
           (cX    , cY    , cZ    , 0.0),
           (nVPNX , nVPNY , nVPNZ , 0.0),
           (0.0   , 0.0   , 0.0   , 1.0))
      end

    fun inverseViewVectorsRotationMatrix (VPN : vec.t, VUP : vec.t) = (* Not tested! *)
      let
        val (nVPNX, nVPNY, nVPNZ) = vec.normalize(VPN);
        val cVUP = vec.crossProduct(VUP, (nVPNX, nVPNY, nVPNZ));
        val (cnVUPX, cnVUPY, cnVUPZ) = vec.normalize(cVUP);
        val (cX, cY, cZ) = vec.crossProduct((nVPNX, nVPNY, nVPNZ),(cnVUPX, cnVUPY, cnVUPZ));
      in
        matr.genericMatrix
          ((cnVUPX, cX , nVPNX, 0.0),
           (cnVUPY, cY , nVPNY, 0.0),
           (cnVUPZ, cZ , nVPNZ, 0.0),
           (0.0   , 0.0, 0.0  , 1.0))
      end

    (* Coorsys mappings *)
    fun fromVectorSysMatrix ((rpX, rpY, rpZ) : vert.vertex, Z : vec.t, UP : vec.t) =
      let
        val rotMatr = inverseViewVectorsRotationMatrix(Z, UP);
      in
        matr.translateXYZ(rotMatr, rpX, rpY, rpZ)
      end

    fun toVectorSysMatrix ((rpX, rpY, rpZ) : vert.vertex, Z : vec.t, UP : vec.t) =
      let
        val transMatr = matr.translateXYZMatrix(~rpX, ~rpY, ~rpZ);
        val rotMatr = viewVectorsRotationMatrix(Z, UP);
      in
        matr.multiply(transMatr, rotMatr)
      end
    
  end
;