signature CE3DCAMERA =
  sig
    structure rect : CERECT;
    structure ang  : CE3DANGLE;
    structure vert : CE3DVERTEX;
    structure edg  : CE3DEDGE;
    structure vec : CE3DVECTOR;
    structure matr : CE3DMATRIX;
    structure poly : CE3DPOLY;
    structure Geom : GEOMETRY;

    datatype projectionType = parallel | perspective;
    
    type camera;

    val defaultCamera         : unit -> camera;
    val experimentalCamera    : unit -> camera;
    val cameraToVRP_VPN_VUP   : camera -> vec.t* vec.t* vec.t;
    val cameraToVMMInput      : camera -> rect.rect * projectionType *
                                          (vert.vertex* real* real);
    val dolly                 : (camera* real) -> camera;
    val tumbleX               : (camera* ang.angle) -> camera;
    val tumbleY               : (camera* ang.angle) -> camera;
    val trackX                : (camera* real) -> camera;
    val trackY                : (camera* real) -> camera;

    val viewOrientationMatrix : (vert.vertex *  (* View reference point (VRP) *)
                                 vec.t *        (* View projection normal (VPN) *)
                                 vec.t)         (* View up vector (VUP) *)
                                 -> matr.t;
    val viewMappingMatrix   : rect.rect ->         (* Viewplane boundaries *)
                              projectionType ->
                              (vert.vertex*        (* Projection Reference Point*)
                               real* real)         (* Front and back clipping planes *)
                              -> (matr.t*
                                  (real* real* real))
                                
    val clip3DLine : (real) -> (vert.vertex * vert.vertex) -> (bool* vert.vertex* vert.vertex);(* FIXME: Should be SOME of (vertex*vertex) *)
    val project3DLineTo2D  : projectionType -> rect.rect -> (vert.vertex * vert.vertex) -> Geom.line;  
  end
;