structure ce3DMatrix : CE3DMATRIX =
  struct
    structure ang : CE3DANGLE = ce3DAngle;
    structure vert : CE3DVERTEX = ce3DVertex;
    structure vec : CE3DVECTOR = ce3DVector;
    structure Math : MATH = Math;
    structure exc : CEEXCEPTIONS = ceExceptions;
    type t = (real*real*real*real)*
             (real*real*real*real)*
             (real*real*real*real)*
             (real*real*real*real);
    exception Divide;(* FIXME: Must be a general exception *)
    
    (* Create an empty matrix of only 0's. (addition identity) *)
    val zeroValue =
      ((0.0, 0.0, 0.0, 0.0),
       (0.0, 0.0, 0.0, 0.0),
       (0.0, 0.0, 0.0, 0.0),
       (0.0, 0.0, 0.0, 0.0))

    (* Create an identity matrix for multiplication *)
    val unitValue =
      ((1.0, 0.0, 0.0, 0.0),
       (0.0, 1.0, 0.0, 0.0),
       (0.0, 0.0, 1.0, 0.0),
       (0.0, 0.0, 0.0, 1.0))
    

    fun genericMatrix (matr : ((real* real* real* real)*
                              (real* real* real* real)*
                              (real* real* real* real)*
                              (real* real* real* real))) =
          (matr : t)

    fun rotateXMatrixSin (sinAngle : real, cosAngle : real) =
      ((1.0, 0.0, 0.0, 0.0),
       (0.0, cosAngle, ~sinAngle, 0.0),
       (0.0, sinAngle, cosAngle, 0.0),
       (0.0, 0.0, 0.0, 1.0))

    fun rotateXMatrix (angle : ang.angle) =
      rotateXMatrixSin(Math.sin(angle), Math.cos(angle))
    
    fun rotateYMatrixSin (sinAngle : real, cosAngle : real) =
      ((cosAngle, 0.0, sinAngle, 0.0),
       (0.0, 1.0, 0.0, 0.0),
       (~sinAngle, 0.0, cosAngle, 0.0),
       (0.0, 0.0, 0.0, 1.0))

    fun rotateYMatrix (angle : ang.angle) =
      rotateYMatrixSin(Math.sin(angle), Math.cos(angle))

    fun rotateZMatrixSin (sinAngle : real, cosAngle : real) =
      ((cosAngle, ~sinAngle, 0.0, 0.0),
       (sinAngle, cosAngle, 0.0, 0.0),
       (0.0, 0.0, 1.0, 0.0),
       (0.0, 0.0, 0.0, 1.0))

    fun rotateZMatrix (angle : ang.angle) =
      rotateZMatrixSin(Math.sin(angle), Math.cos(angle))

    fun scaleXYZMatrix (scaleX : real, scaleY : real, scaleZ : real) =
      ((scaleX, 0.0, 0.0, 0.0),
       (0.0, scaleY, 0.0, 0.0),
       (0.0, 0.0, scaleZ, 0.0),
       (0.0, 0.0, 0.0, 1.0))
      
    fun translateXYZMatrix (translateX : real, translateY : real, translateZ : real) =
      ((1.0, 0.0, 0.0, translateX),
       (0.0, 1.0, 0.0, translateY),
       (0.0, 0.0, 1.0, translateZ),
       (0.0, 0.0, 0.0, 1.0))

    fun shearXYMatrix (shearX : real, shearY : real) =
      ((1.0, 0.0, shearX, 0.0),
       (0.0, 1.0, shearY, 0.0),
       (0.0, 0.0, 1.0, 0.0),
       (0.0, 0.0, 0.0, 1.0))
      
    fun shearXZMatrix (shearX : real, shearZ : real) =
      ((1.0, shearX, 0.0, 0.0),
       (0.0, 1.0, 0.0, 0.0),
       (0.0, shearZ, 1.0, 0.0),
       (0.0, 0.0, 0.0, 1.0))
    
    fun shearYZMatrix (shearY : real, shearZ : real) =
      ((1.0, 0.0, 0.0, 0.0),
       (shearY, 1.0, 0.0, 0.0),
       (shearZ, 0.0, 1.0, 0.0),
       (0.0, 0.0, 0.0, 1.0))

    fun swapLRCoordsysMatrix () = (* Not tested! *)
      ((1.0, 0.0, 0.0, 0.0),
       (0.0, 1.0, 0.0, 0.0),
       (0.0, 0.0, ~1.0, 0.0),
       (0.0, 0.0, 0.0, 1.0))

    fun multiply (matr1 : t,matr2 : t) =
      let
        val (mr0,mr1,mr2,mr3) = matr1;
        val (mrr0,mrr1,mrr2,mrr3) = matr2;
        fun getIndex ((in0,in1,in2,in3), 0) = in0
          | getIndex ((in0,in1,in2,in3), 1) = in1
          | getIndex ((in0,in1,in2,in3), 2) = in2
          | getIndex ((in0,in1,in2,in3), 3) = in3
          | getIndex (_,_) = raise ceExceptions.generic("fatal error in GetIndex! Index outside 0-4!")
          
        fun multRowCol ((in0,in1,in2,in3), index) =
          ((getIndex(mr0,index) * in0) +
           (getIndex(mr1,index) * in1) +
           (getIndex(mr2,index) * in2) +
           (getIndex(mr3,index) * in3))

      in
        ((multRowCol(mrr0,0),multRowCol(mrr0,1),multRowCol(mrr0,2),multRowCol(mrr0,3)),
         (multRowCol(mrr1,0),multRowCol(mrr1,1),multRowCol(mrr1,2),multRowCol(mrr1,3)),
         (multRowCol(mrr2,0),multRowCol(mrr2,1),multRowCol(mrr2,2),multRowCol(mrr2,3)),
         (multRowCol(mrr3,0),multRowCol(mrr3,1),multRowCol(mrr3,2),multRowCol(mrr3,3)))
      end
    ;

    fun divide (a : t, b : t) =
      if true then
        raise exc.notImplemented (* FIXME: Not implemented... *)
      else
        a
    
    fun reciprocal (a : t) =
      if true then
        raise exc.notImplemented (* FIXME: Not implemented... *)
      else
        a

    fun add ((amr0,amr1,amr2,amr3) : t,
             (bmr0,bmr1,bmr2,bmr3) : t) = (* Not tested! *)
      let
        fun addRows((ac0,ac1,ac2,ac3), (bc0,bc1,bc2,bc3)) =
          (ac0+bc0, ac1+bc1, ac2+bc2, ac3+bc3)
      in
        (addRows(amr0,bmr0),
         addRows(amr1,bmr1),
         addRows(amr2,bmr2),
         addRows(amr3,bmr3))
      end
    
    fun subtract ((amr0,amr1,amr2,amr3) : t,
                  (bmr0,bmr1,bmr2,bmr3) : t) = (* Not tested! *)
      let
        fun subtractRows((ac0,ac1,ac2,ac3), (bc0,bc1,bc2,bc3)) =
          (ac0-bc0, ac1-bc1, ac2-bc2, ac3-bc3)
      in
        (subtractRows(amr0,bmr0),
         subtractRows(amr1,bmr1),
         subtractRows(amr2,bmr2),
         subtractRows(amr3,bmr3))
      end
    
    fun negate ((mr0, mr1, mr2, mr3) : t) = (* Not tested! *)
      let
        fun negateRow(c0,c1,c2,c3) =
          (~c0, ~c1, ~c2, ~c3)
      in
        (negateRow(mr0),
         negateRow(mr1),
         negateRow(mr2),
         negateRow(mr3))
      end
    
    fun scalarMultiply ((mr0, mr1, mr2, mr3) : t, factor) = (* Not tested! *)
      let
        fun multRow(c0,c1,c2,c3) =
          (c0 * factor, c1 * factor, c2 * factor, c3 * factor)
      in
        (multRow(mr0),
         multRow(mr1),
         multRow(mr2),
         multRow(mr3))
      end

    fun scalarDivide ((mr0, mr1, mr2, mr3) : t, divisor) = (* Not tested! *)
      let
        fun divRow(c0,c1,c2,c3) =
          (c0 / divisor, c1 / divisor, c2 / divisor, c3 / divisor)
      in
        if Real.== (divisor, 0.0) then
          raise Divide
        else
          (divRow(mr0),
           divRow(mr1),
           divRow(mr2),
           divRow(mr3))
      end
    
    fun equal ((amr0,amr1,amr2,amr3) : t,
               (bmr0,bmr1,bmr2,bmr3) : t) = (* Not tested! *)
      let
        fun compareRows((ac0,ac1,ac2,ac3), (bc0,bc1,bc2,bc3)) =
          (Real.== (ac0, bc0) andalso
           Real.== (ac1, bc1) andalso
           Real.== (ac2, bc2) andalso
           Real.== (ac3, bc3))
      in
        (compareRows(amr0,bmr0) andalso
         compareRows(amr1,bmr1) andalso
         compareRows(amr2,bmr2) andalso
         compareRows(amr3,bmr3))
      end
    
    fun scaleXYZ (matr : t,scaleX : real,scaleY : real,scaleZ : real) =
      let
        val ((in00,in01,in02,in03),
             (in10,in11,in12,in13),
             (in20,in21,in22,in23),row3) = matr;
      in
        ((in00*scaleX,in01*scaleX,in02*scaleX,in03*scaleX),
         (in10*scaleY,in11*scaleY,in12*scaleY,in13*scaleY),
         (in20*scaleZ,in21*scaleZ,in22*scaleZ,in23*scaleZ),
         row3)
      end
    ;
    fun rotateXsin (matr : t,sinAngle : real, cosAngle : real) =
      let
        val (row0,(in10,in11,in12,in13),(in20,in21,in22,in23),row3) = matr;
      in
        (row0,
         ((in10*cosAngle)-(in20*sinAngle),
          (in11*cosAngle)-(in21*sinAngle),
          (in12*cosAngle)-(in22*sinAngle),
          (in13*cosAngle)-(in23*sinAngle)),
         ((in10*sinAngle)+(in20*cosAngle),
          (in11*sinAngle)+(in21*cosAngle),
          (in12*sinAngle)+(in22*cosAngle),
          (in13*sinAngle)+(in23*cosAngle)),
         row3)
      end
    ;
    fun rotateX (matr : t,angle : ang.angle) =
      rotateXsin (matr,Math.sin(angle),Math.cos(angle))
    ;
    fun rotateYsin (matr : t,sinAngle : real, cosAngle : real) =
      let
        val ((in00,in01,in02,in03),row1,(in20,in21,in22,in23),row3) = matr;
      in
        (((in00*cosAngle)+(in20*sinAngle),
          (in01*cosAngle)+(in21*sinAngle),
          (in02*cosAngle)+(in22*sinAngle),
          (in03*cosAngle)+(in23*sinAngle)),
         row1,
         ((in20*cosAngle)-(in00*sinAngle),
          (in21*cosAngle)-(in01*sinAngle),
          (in22*cosAngle)-(in02*sinAngle),
          (in23*cosAngle)-(in03*sinAngle)),
         row3)
      end
    ;
    fun rotateY (matr : t,angle : ang.angle) =
      rotateYsin (matr,Math.sin(angle),Math.cos(angle))
    ;
    fun rotateZsin(matr : t,sinAngle : real,cosAngle : real) =
      let
        val ((in00,in01,in02,in03),(in10,in11,in12,in13),row2,row3) = matr;
      in
        (((in00*cosAngle)-(in10*sinAngle),
          (in01*cosAngle)-(in11*sinAngle),
          (in02*cosAngle)-(in12*sinAngle),
          (in03*cosAngle)-(in13*sinAngle)),
         ((in00*sinAngle)+(in10*cosAngle),
          (in01*sinAngle)+(in11*cosAngle),
          (in02*sinAngle)+(in12*cosAngle),
          (in03*sinAngle)+(in13*cosAngle)),
         row2,
         row3)
      end
    ;
    fun rotateZ (matr : t,angle : ang.angle) =
      rotateZsin (matr,Math.sin(angle),Math.cos(angle))
    ;
    fun translateXYZ (matr : t,translateX : real,translateY : real,translateZ : real) =
      let
        val ((in00,in01,in02,in03),(in10,in11,in12,in13),(in20,in21,in22,in23),row3) = matr;
      in
        (* This only works if row3 is always [0,0,0,1]!!! *)
        ((in00,in01,in02,in03+translateX),
         (in10,in11,in12,in13+translateY),
         (in20,in21,in22,in23+translateZ),
         row3)
      end
    ;

    fun shearXY (matr : t ,shearX : real, shearY : real) =
      let
        val ((in00,in01,in02,in03),(in10,in11,in12,in13),(in20,in21,in22,in23),row3) = matr;
      in
        ((in00 + (in20 * shearX),in01 + (in21 * shearX), in02 + (in22 * shearX), in03 + (in23 * shearX)),
         (in10 + (in20 * shearY),in11 + (in21 * shearY), in12 + (in22 * shearY), in13 + (in23 * shearY)),
         (in20,in21,in22,in23),
         row3)
      end
    ;

    fun shearXZ (matr : t ,shearX : real, shearZ : real) =
      let
        val ((in00,in01,in02,in03),(in10,in11,in12,in13),(in20,in21,in22,in23),row3) = matr;
      in
        ((in00 + (in10 * shearX),in01 + (in11 * shearX), in02 + (in12 * shearX), in03 + (in13 * shearX)),
         (in10,in11,in12,in13),
         (in20 + (in10 * shearZ),in21 + (in11 * shearZ), in22 + (in12 * shearZ), in23 + (in13 * shearZ)),
         row3)
      end
    ;

    fun shearYZ (matr : t ,shearY : real, shearZ : real) =
      let
        val ((in00,in01,in02,in03),(in10,in11,in12,in13),(in20,in21,in22,in23),row3) = matr;
      in
        ((in00,in01,in02,in03),
         (in10 + (in00 * shearY),in11 + (in01 * shearY), in12 + (in02 * shearY), in13 + (in03 * shearY)),
         (in20 + (in00 * shearZ),in21 + (in01 * shearZ), in22 + (in02 * shearZ), in23 + (in03 * shearZ)),
         row3)
      end
    ;
    
    fun swapLRCoordsys (matr : t) =
      let
        val (row0,row1,(in20,in21,in22,in23),row3) = matr;
      in
        (row0,row1,(~in20,~in21,~in22,~in23),row3)
      end
    ;
    
    (* Return a function that transforms one 3D-vertex (ce3DVertex) *)
    fun transformVertexFunction (matr : t) (cX,cY,cZ) =
      let
        val ((in00,in01,in02,in03),
             (in10,in11,in12,in13),
             (in20,in21,in22,in23),
             _) = matr;
      in
        ((in00*cX)+(in01*cY)+(in02*cZ)+in03,
         (in10*cX)+(in11*cY)+(in12*cZ)+in13,
         (in20*cX)+(in21*cY)+(in22*cZ)+in23)
      end
    ;
  end

structure ceTest3DMatrix : CETESTER =
  struct
    structure matr = ce3DMatrix;
    
    fun translate () = 
          matr.equal(
            matr.translateXYZMatrix(1.6, 6.1, ~7.2),
            matr.translateXYZ(matr.unitValue ,1.6, 6.1, ~7.2))

    fun test () =
          ceTU.doTest("translate", translate)
  end