structure ce3DPrimitiveGenerators : CE3DPRIMITIVEGENERATORS =
  struct
    structure arr : ARRAY = Array;
    structure vert : CE3DVERTEX = ce3DVertex;
    structure edg : CE3DEDGE = ce3DEdge;
    structure matr : CE3DMATRIX = ce3DMatrix;
    structure ang  : CE3DANGLE = ce3DAngle;
    structure poly : CE3DPOLY = ce3DPoly;

    fun generateSphere (uSections : int,vSections : int,radius : real) =
      if (uSections < 4) then
        raise Fail "uSections is < 4 in fun generateSphere()"
     else if(vSections < 4) then
       raise ceExceptions.generic "vSections is < 4 in generateSphere()"
     else
       let
         fun calcV(vertexList,u:int,v:int) =
           if v<vSections then
             let
               (* FIXME: I know that this is the slow way of doing it... *)
               val matrix = matr.unitValue;
               val matrix = matr.rotateZ(matrix,(ang.deg180*real(u))/(real(uSections-1)) : ang.angle);
               val matrix = matr.rotateY(matrix,(ang.deg360*real(v))/real(vSections) : ang.angle);
               val transFun = matr.transformVertexFunction(matrix);
             in
               calcV(transFun(0.0, radius, 0.0)::vertexList,u,v+1)
             end
           else
             vertexList
       
         fun calcU(vertexList,u:int) =
           if u<(uSections-1) then
             calcU(calcV(vertexList,u,0),u+1)
           else
             vertexList

         val (topVertex, bottomVertex) = ((0.0, radius, 0.0), (0.0, ~radius, 0.0));
         val vertexList = bottomVertex::calcU([topVertex],1);

         fun endEdges(edgeList, v : int, endVert : int, firstVert : int) =
           if v < vSections then
             (endVert,v+firstVert,~1,~1)::endEdges(edgeList,v+1, endVert, firstVert)
           else
             edgeList

         fun sideEdges(edgeList, startIndex : int) =
           let
             fun circleEdges(innerEdgeList, v : int) =
               if (v < vSections) then
                 let
                   val fromVertex = startIndex + v;
                   val toVertex = startIndex +
                                    (if (v = (vSections - 1)) then
                                       (0)
                                     else
                                       (v + 1));
                   fun addPerpenEdge(list) =
                     if (startIndex < (vSections*(uSections - 3))) then
                       ((fromVertex, fromVertex + vSections, ~1, ~1)::list)
                     else
                       (list)
                 in
                   circleEdges( addPerpenEdge((fromVertex, toVertex, ~1, ~1)::innerEdgeList), v + 1)
                 end
               else
                 innerEdgeList
           in
             if (startIndex < (vSections*(uSections - 2))) then
               sideEdges(circleEdges(edgeList, 0), startIndex + vSections)
             else
               edgeList
           end

         val topEndEdges = endEdges(nil, 0, 0, 1);
         val edgeList = sideEdges(topEndEdges, 1);
         val edgeList = endEdges(edgeList, 0, (vSections*(uSections-2)) + 1, (vSections*(uSections-3))+1);
         val polygonList = nil;
       in
         (Array.fromList(vertexList), Array.fromList(edgeList), Array.fromList(polygonList))
       end

    fun generateGrid(xSize : int, ySize : int) =
          if (xSize < 1) then
            raise ceExceptions.generic("xSize < 1 in generateGrid()")
          else if (ySize < 1) then
            raise ceExceptions.generic("ySize < 1 in generateGrid()")
          else
            let
              fun vertGen(~1) =
                    let
                      fun genY(~1) =
                             nil
                        | genY(y) =
                             (Real.fromInt(~xSize), 0.0, Real.fromInt(~ySize + (y*2)))::
                             (Real.fromInt(xSize),  0.0, Real.fromInt(~ySize + (y*2)))::
                             genY(y - 1)
                    in
                      genY(ySize)
                    end
                | vertGen(x) =
                    (Real.fromInt(~xSize + (x*2)), 0.0, Real.fromInt(~ySize))::
                    (Real.fromInt(~xSize + (x*2)), 0.0, Real.fromInt(ySize) )::
                    vertGen(x - 1)

              fun edgeGen (~1) =
                    nil
                | edgeGen (cnt) =
                    (cnt, cnt - 1, ~1, ~1)::edgeGen(cnt-2)
            in
              (Array.fromList(vertGen(xSize)) : poly.vertexArray,
               Array.fromList(edgeGen((((xSize+1)+(ySize+1))*2) - 1)) : poly.edgeArray,
               Array.fromList(nil) : poly.polygonArray)
      end
    
  end
;
