structure ce3DVector : CE3DVECTOR =
  struct
    structure Math : MATH = Math;

    type vertex = real* real* real;
    type t = vertex;
    exception Divide;(* FIXME: Should be a general exception *)

    (* Methods for addition arithmetics *)
    val zeroValue =
          (0.0, 0.0, 0.0)
    fun negate ((x,y,z) : t) =
          (~x, ~y, ~z)
    fun add ((x0, y0, z0) : t,(x1, y1, z1) : t) =
          (x0 + x1, y0 + y1, z0 + z1)
    fun subtract ((x0, y0, z0) : t,(x1, y1, z1) : t) =
          (x0 - x1, y0 - y1, z0 - z1)
    fun scalarMultiply ((x, y, z) : t, factor : real) =
          (x* factor, y* factor, z* factor)
    fun scalarDivide ((x, y, z) : t, divisor : real) =
          if Real.==(0.0, divisor) then
            raise Divide
          else
            (x / divisor, y / divisor, z / divisor)
    fun equal ((x0, y0, z0) : t, (x1, y1, z1) : t) =
          (Real.==(x0, x1) andalso
           Real.==(y0, y1) andalso
           Real.==(z0, z1))

    (* Special 3D matrix methods... *)
    fun fromVertices((x0, y0, z0) : vertex,(x1, y1, z1) : vertex) =
          (x1-x0, y1-y0, z1-z0)
    fun dotProduct((x0,y0,z0) : t,(x1,y1,z1) : t) =
          ((x0*x1)+(y0*y1)+(z0*z1))
    fun crossProduct((x0,y0,z0) : t,(x1,y1,z1) : t) =
          (((y0*z1)-(z0*y1)),((z0*x1)-(x0*z1)),((x0*y1)-(y0*x1)))
    fun vectorLength ((x,y,z) : t) =
          Math.sqrt((x*x)+(y*y)+(z*z));
    fun normalize ((x,y,z) : t) =
          let
            val vecLen = vectorLength(x,y,z);
          in
            if Real.==(vecLen, 0.0) then
              raise Divide
            else
              (x/vecLen,y/vecLen,z/vecLen)
          end
  end
;