structure ceComplexRect : CECOMPLEXRECT = (* Structure of complex numbers in rectangular representation *)
  struct
    type t = real* real;(* (realPart, imagPart) *)
    exception Divide;(* FIXME: Should be a general exception *) 

    (* Methods for addition arithmetics *)
    val zeroValue =
          ((0.0, 0.0) : t)
    fun add ((real0, imag0) : t, (real1, imag1) : t) =
          (real0 + real1, imag0 + imag1)
    fun subtract ((real0, imag0) : t, (real1, imag1) : t) =
          (real0 - real1, imag0 - imag1)
    fun negate ((realPart, imagPart) : t) =
          (~realPart, ~imagPart)
    fun scalarMultiply ((real0, imag0) : t, factor) =
          (real0 * factor, imag0 * factor)
    fun scalarDivide ((real0, imag0) : t, divisor) =
          if Real.==(0.0, divisor) then
            raise Divide
          else
            (real0 / divisor, imag0 / divisor)
    fun equal ((real0, imag0) : t, (real1, imag1) : t) =
          (Real.==(real0, real1) andalso Real.==(imag0, imag1))

    (* Methods for multiplication arithmetics *)
    val unitValue =
          ((1.0, 0.0) : t)
    fun multiply ((real0, imag0) : t, (real1, imag1) : t) =
          (real0 * real1 - imag0 * imag1, real0 * imag1 + real1 * imag0)
    fun divide ((real0, imag0) : t, (real1, imag1) : t) =
          let
            val denom = real1 * real1 + imag1 * imag1;
          in
            if Real.== (denom, 0.0) then
              raise Divide
            else
              ((real0 * real1 + imag0 * imag1) / denom,
               (imag0 * real1 - real0 * imag1) / denom)
          end
    fun reciprocal ((real0, imag0) : t) =
          let (* As divide ((1.0, 0) / (real0, imag0)) *)
            val denom = real0 * real0 + imag0 * imag0;
          in
            if Real.== (denom, 0.0) then
              raise Divide
            else
              (real0 / denom, ~(imag0 / denom))
          end

    (* Methods specific to complex numbers in rect. rep. *)
    fun fromRectangular (realPart, imagPart) =
          ((realPart, imagPart) : t)
    (*val fromRectangular : {real: real, imag: real} -> t;*)
    (*val fromComplexPol  : ceComplexPolar.t *)
    fun realPart (realPar, imagPar) =
          (realPar)
    fun imagPart (realPar, imagPar) =
          (imagPar)
    fun realAndImag (realPar, imagPar) =
          (realPar, imagPar)

    fun konjug ((real0, imag0) : t) =
          (real0, ~imag0)
  end
;