structure ceWireRender : CERENDER =
  struct
    structure rect : CERECT = ceRect;
    structure ang  : CE3DANGLE = ce3DAngle;
    structure vert : CE3DVERTEX = ce3DVertex;
    structure edg  : CE3DEDGE = ce3DEdge;
    structure vec : CE3DVECTOR = ce3DVector;
    structure matr : CE3DMATRIX = ce3DMatrix;
    structure amatr : CE3DADVMATRIX = ce3DAdvMatrix;
    structure poly : CE3DPOLY = ce3DPoly;
    structure Geom : GEOMETRY = Geometry;
    structure cam  : CE3DCAMERA = ce3DCamera;
    structure math : MATH = Math;

    fun render(camera,
               viewPort : rect.rect,
               mesh : poly.polygonMesh) =
      let
        val voMatrix = cam.viewOrientationMatrix(
                         cam.cameraToVRP_VPN_VUP(camera));
        val (screenWindow, camType, projInfo) = cam.cameraToVMMInput(camera);
        val (vmMatrix, (zProj, zMin, zMax)) =
              cam.viewMappingMatrix screenWindow camType projInfo;
        val viewMatrix = matr.multiply(voMatrix, vmMatrix);

        val transformFnc = matr.transformVertexFunction viewMatrix;
        fun foldTFunc (elem, list) =
              (transformFnc elem)::list
        val (meshVert, meshEdge, meshPoly) = mesh;
        val transVert =
              Array.fromList
                (Array.foldr foldTFunc nil meshVert);

        val clip3D = cam.clip3DLine zMin;
        val proj3DTo2D = cam.project3DLineTo2D camType viewPort;

        fun foldGenLines ((sp, ep, _, _), list) =
          let
            val (OK, startPoint, endPoint) =
                  clip3D(Array.sub(transVert, sp),
                         Array.sub(transVert, ep));
          in
            if OK then
              (proj3DTo2D(startPoint, endPoint))::list
            else
              list
          end
      in
        (Array.foldr foldGenLines nil meshEdge)
      end
  end
;